/* 
 * glass - the openGL Articulated Skeleton System
 *
 * Copyright (C) 2001-2003  Robert Ancell <bob27@users.sourceforge.net>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include <stdlib.h>

#include "global.h"
#include "glass_draw.h"

static void _glassSetMaterial(GlassMaterial *material);
static void _glassDrawTriangle(GlassTriangle *triangle);
static void _glassDrawAlphaTriangle(GlassTriangle *alpha_triangle);

/* Draws the glass model */
int _glassDrawModel(GlassModel *model)
{
  if(model->children == NULL)
    return TRUE; /* empty model !!should return an error? */

  /* All textures drawn in modulate mode */
  glTexEnvf(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_MODULATE);

  /* Draw the nodes of the model */
  /* Normal triangles */
  glPushMatrix();
  if(!_glassDrawNode(model->children))
    return FALSE;
  glPopMatrix();

  /* Alpha Triangles */
  return _glassDrawAlphaNode(model->children);
}

/* Recursively draw the normal triangles of a node */
/* FIXME: doesn't check to see how many time glPushMatrix is used, errors */
int _glassDrawNode(GlassNode *node)
{
  if(node == NULL)
    return TRUE;

  glPushMatrix();

  /* Do a transform */
  _glassDoTransform(node->transform);

  /* Do a child node (will inherit this transform) */
  _glassDrawNode(node->children);

  /* Draw the surface associated with that transform */
  if(node->surface && node->surface->visible)
    _glassDrawSurface(node->surface);

  glPopMatrix();

  /* Move to the next transform */
  _glassDrawNode(node->next);

  return TRUE;
}

/* Recursively draw the alpha triangles of a node */
/* FIXME: doesn't check to see how many time glPushMatrix is used, errors */
int _glassDrawAlphaNode(GlassNode *node)
{
  glPushMatrix();

  /* Do a transform */
  _glassDoTransform(node->transform);

  /* Do inherited transform */
  if(node->children)
    _glassDrawAlphaNode(node->children);

  /* Draw the surface associated with that transform */
  if(node->surface != NULL && node->surface->visible)
    _glassDrawAlphaSurface(node->surface);

  glPopMatrix();

  /* Move to the next transform */
  if(node->next != NULL)
    _glassDrawAlphaNode(node->next);

  return TRUE;
}

/* Convert the transform in the glass model to openGL calls */
int _glassDoTransform(GlassTransform *transform)
{
  GlassTransformType type;
  GLfloat **values;

  if(transform == NULL)
    return TRUE;

  type = transform->type;
  if(type == GLASS_NONE)
    return TRUE;

  values = transform->values; /* The array of values */

  /* Do the transform */
  switch(type)
  {
  case GLASS_TRANSLATE:
    glTranslatef(*values[0], *values[1], *values[2]);
    break;
  case GLASS_ROTATE:
    glRotatef(*values[0], *values[1], *values[2], *values[3]);
    break;
  case GLASS_SCALE:
    glScalef(*values[0], *values[1], *values[2]);
    break;
  case GLASS_TRANSLATE_X:
    glTranslatef(*values[0], 0.0f, 0.0f);
    break;
  case GLASS_TRANSLATE_Y:
    glTranslatef(0.0f, *values[0], 0.0f);
    break;
  case GLASS_TRANSLATE_Z:
    glTranslatef(0.0f, 0.0f, *values[0]);
    break;
  case GLASS_ROTATE_X:
    glRotatef(*values[0], 1.0f, 0.0f, 0.0f);
    break;
  case GLASS_ROTATE_Y:
    glRotatef(*values[0], 0.0f, 1.0f, 0.0f);
    break;
  case GLASS_ROTATE_Z:
    glRotatef(*values[0], 0.0f, 0.0f, 1.0f);
    break;
  case GLASS_SCALE_X:
    glScalef(*values[0], 1.0f, 1.0f);
    break;
  case GLASS_SCALE_Y:
    glScalef(1.0f, *values[0], 1.0f);
    break;
  case GLASS_SCALE_Z:
    glScalef(1.0f, 1.0f, *values[0]);
    break;
  case GLASS_SCALE_EVEN:
    glScalef(*values[0], *values[0], *values[0]);
    break;
  default:
    return FALSE;
    break;
  }

  /* stop the normal (non)inversion */
  /*if((type >= GLASS_SCALE_X) && (type <= GLASS_SCALE_Z))
    glCullFace(GL_BACK);
  else
    glCullFace(GL_FRONT);*/

  return TRUE;
}

/* Draw the non-alpha triangles in a surface */
int _glassDrawSurface(GlassSurface *surface)
{
  GlassTriangle *triangle = surface->triangles; /* The surface of triangles */
  GlassMaterial *current_material = NULL;

  /* If a surface has been created */
  if(surface->index)
  {
    glCallList(surface->index);
    return TRUE;
  }

  /* Set to draw opaque triangles */
  glPushAttrib(GL_ENABLE_BIT | GL_DEPTH_BUFFER_BIT | GL_COLOR_BUFFER_BIT);
  glDisable(GL_BLEND);
  glDepthMask(GL_TRUE);

  /* Set the default material */
  _glassSetMaterial(NULL);

  glBegin(GL_TRIANGLES);
  /* Otherwise draw the triangles manually */
  while(triangle != NULL)
  {
    if(triangle->material != current_material)
    {
      glEnd();
      _glassSetMaterial(triangle->material);
      glBegin(GL_TRIANGLES);
      current_material = triangle->material;
    }

    _glassDrawTriangle(triangle);
    triangle = triangle->next;
  }
  glEnd();

  glPopAttrib();

  return TRUE;
}

/* Draw the alpha triangles in a surface */
int _glassDrawAlphaSurface(GlassSurface *surface)
{
  GlassTriangle *triangle = surface->triangles; /* The surface of triangles */
  GlassMaterial *current_material = NULL;

  /* If a surface has been created */
  if(surface->alpha_index)
  {
    glCallList(surface->alpha_index);
    return TRUE;
  }

  /* Set to draw transparent triangles */
  glPushAttrib(GL_ENABLE_BIT | GL_DEPTH_BUFFER_BIT | GL_COLOR_BUFFER_BIT);
  glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
  glEnable(GL_BLEND);
  glDepthMask(GL_FALSE);

  /* Set the default material */
  _glassSetMaterial(NULL);

  /* Otherwise draw the triangles manually */
  glBegin(GL_TRIANGLES);
  triangle = surface->alpha_triangles;
  while(triangle != NULL)
  {
    if(triangle->material != current_material)
    {
      glEnd();
      _glassSetMaterial(triangle->material);
      glBegin(GL_TRIANGLES);
      current_material = triangle->material;
    }

    _glassDrawAlphaTriangle(triangle);
    triangle = triangle->next;
  }
  glEnd();

  glPopAttrib();

  return TRUE;
}

/* Change the current material properties */
static void _glassSetMaterial(GlassMaterial *material)
{
  if(material == NULL)
  {
    GLfloat lgrey[4] = {0.8, 0.8, 0.8, 1.0}, dgrey[4] = {0.2, 0.2, 0.2, 1.0};
    GLfloat black[4] = {0.0, 0.0, 0.0, 1.0};

    /* Set some defaults */
    glMaterialfv(GL_FRONT_AND_BACK, GL_AMBIENT, dgrey);
    glMaterialfv(GL_FRONT_AND_BACK, GL_DIFFUSE, lgrey);
    glMaterialfv(GL_FRONT_AND_BACK, GL_SPECULAR, black);
    glMaterialfv(GL_FRONT_AND_BACK, GL_EMISSION, black);
    glMaterialf(GL_FRONT_AND_BACK, GL_SHININESS, 0.0);

    glDisable(GL_TEXTURE_2D);

    return;
  }

  glMaterialfv(GL_FRONT_AND_BACK, GL_AMBIENT, material->ambient);
  glMaterialfv(GL_FRONT_AND_BACK, GL_DIFFUSE, material->diffuse);
  glMaterialfv(GL_FRONT_AND_BACK, GL_SPECULAR, material->specular);
  glMaterialfv(GL_FRONT_AND_BACK, GL_EMISSION, material->emission);
  glMaterialf(GL_FRONT_AND_BACK, GL_SHININESS, material->shininess);

  if(material->texture)
  {
    glEnable(GL_TEXTURE_2D);
    glBindTexture(GL_TEXTURE_2D, material->texture->texture);
  }
  else
    glDisable(GL_TEXTURE_2D);
}

/* Draw one triangle */
static void _glassDrawTriangle(GlassTriangle *triangle)
{
  int i;

  for(i = 0; i < 3; i++)
  {
    glNormal3fv(triangle->nor[i]);
    glColor3fv(triangle->col[i]);
    glTexCoord2fv(triangle->tex[i]);
    glVertex3fv(triangle->pos[i]);
  }
}

/* Draw one alpha triangle */
void _glassDrawAlphaTriangle(GlassTriangle *alpha_triangle)
{
  int i;

  for(i = 0; i < 3; i++)
  {
    glNormal3fv(alpha_triangle->nor[i]);
    glColor4fv(alpha_triangle->col[i]);
    glVertex3fv(alpha_triangle->pos[i]);
    glTexCoord2fv(alpha_triangle->tex[i]);
  }
}
