/* 
 * glass - the openGL Articulated Skeleton System
 *
 * Copyright (C) 2001-2003  Robert Ancell <bob27@users.sourceforge.net>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

/* This file contains the functions used externally to access the library */

#include <stdlib.h>

#include "global.h"
#include "glass_load.h"
#include "glass_library.h"
#include "glass_variable.h"
#include "glass_apoint.h"
#include "glass_draw.h"
#include "glass_surface.h"

/* Return a string for this version of glass */
const char *glassGetVersion(void)
{
  static char version[] = VERSION_STRING;

  return version;
}

/* Load an model */
GlassModel *glassLoadModel(const char *fname)
{
  return _glassLoadModel(NULL, fname);
}

/* Make a new library */
GlassLibrary *glassLibraryNew(const char *name)
{
  return _glassLibraryNew(name);
}

/* Load an model into a library */
GlassModel *glassLibraryLoadModel(GlassLibrary *library, const char *fname)
{
  return _glassLoadModel(library, fname);
}

/* Return the number of variables in this model */
int glassGetNumVariables(GlassModel *model)
{
  return model->nvariables;
}

/* Return the index of a variable with the name `name', return -1 if not found */
int glassFindVariable(GlassModel *model, const char *name)
{
  GlassVariable *variable = _glassModelFindVariable(model, name);
  int i;

  if(variable == NULL)
    return -1;

  /* Find the index */
  for(i = 0; i < model->nvariables && model->variable_array[i] != variable; i++);

  return i;
}

/* Get the current value of a variable */
GLfloat glassGetVariableValue(GlassModel *model, int variable)
{
  if(variable >= model->nvariables || variable < 0)
    return FALSE;

  return _glassVariableGetValue(model->variable_array[variable]);
}

/* Get a pointer to the current value of a variable */
const GLfloat *glassGetVariableValuev(GlassModel *model, int variable)
{
  if(variable >= model->nvariables || variable < 0)
    return FALSE;

  return _glassVariableGetValuev(model->variable_array[variable]);
}

/* Get the minimum value a variable can take */
GLfloat glassGetVariableMin(GlassModel *model, int variable)
{
  if(variable >= model->nvariables || variable < 0)
    return FALSE;

  return _glassVariableGetMin(model->variable_array[variable]);
}

/* Get the maximum value a variable can take */
GLfloat glassGetVariableMax(GlassModel *model, int variable)
{
  if(variable >= model->nvariables || variable < 0)
    return FALSE;

  return _glassVariableGetMax(model->variable_array[variable]);
}

/* Set a variable, return FALSE if the variable doesn't exist */
int glassSetVariable(GlassModel *model, int variable, GLfloat value)
{
  if(variable >= model->nvariables || variable < 0)
    return FALSE;

  _glassVariableSetValue(model->variable_array[variable], value);
  model->variables_changed = TRUE;

  return TRUE;
}

/* Increment a variable, return FALSE if the variable doesn't exist */
int glassIncVariable(GlassModel *model, int variable, GLfloat dvalue)
{
  if(variable >= model->nvariables || variable < 0)
    return FALSE;

  _glassVariableIncValue(model->variable_array[variable], dvalue);
  model->variables_changed = TRUE;

  return TRUE;
}

/* Return the number of active points in this model */
int glassGetNumActivePoints(GlassModel *model)
{
  return model->npoints;
}

/* Find an active point */
int glassFindActivePoint(GlassModel *model, const char *name)
{
  GlassActivePoint *apoint = _glassModelFindActivePoint(model, name);
  int i;

  if(apoint == NULL)
    return -1;

  /* Find the index */
  for(i = 0; i < model->npoints && model->apoint_array[i] != apoint; i++);

  return i;
}

/* Get the position of an active point */
GLfloat *glassGetActivePointPos(GlassModel *model, int apoint)
{
  if(apoint >= model->npoints || apoint < 0)
    return NULL;

  /* Update the active points if necessary */
  if(model->variables_changed)
  {
    _glassModelUpdateActivePoints(model);
    model->variables_changed = FALSE;
  }

  return _glassActivePointGetPos(model->apoint_array[apoint]);
}

/* Get the direction of an active point */
GLfloat *glassGetActivePointDir(GlassModel *model, int apoint)
{
  if(apoint >= model->npoints || apoint < 0)
    return NULL;

  /* Update the active points if necessary */
  if(model->variables_changed)
  {
    _glassModelUpdateActivePoints(model);
    model->variables_changed = FALSE;
  }

  return _glassActivePointGetDir(model->apoint_array[apoint]);
}

/* Draw the model, return FALSE if error occurs */
int glassDrawModel(GlassModel *model)
{
   /* Update the active points if necessary */
  if(model->variables_changed)
  {
    _glassModelUpdateActivePoints(model);
    model->variables_changed = FALSE;
  }
   
  return _glassDrawModel(model);
}

/* Return the number of surfaces in this model */
int glassGetNumSurfaces(GlassModel *model)
{
  return model->nsurfaces;
}

/* Return the index of a surface with the name `name', return -1 if not found */
int glassFindSurface(GlassModel *model, const char *name)
{
  GlassSurface *surface = _glassModelFindSurface(model, name);
  int i;

  if(surface == NULL)
    return -1;

  /* Find the index */
  for(i = 0; i < model->nsurfaces && model->surface_array[i] != surface; i++);

  return i;
}

/* Toggle whether or not this surface is drawn or not,
   visible = TRUE = draw surface, visible = FALSE = hide surface.
   Return FALSE if surface doesn't exist */
int glassSetSurfaceVisibility(GlassModel *model, int surface, int visible)
{
  if(surface >= model->nsurfaces || surface < 0)
    return FALSE;

  _glassSurfaceSetVisibility(model->surface_array[surface], visible);
  return TRUE;
}

/* Draw one surface of an model, return FALSE if surface doesn't exist */
int glassDrawSurface(GlassModel *model, int surface)
{
  if(surface >= model->nsurfaces || surface < 0)
    return FALSE;

  if(!_glassDrawSurface(model->surface_array[surface]))
    return FALSE;
  return _glassDrawAlphaSurface(model->surface_array[surface]);
}

/* Remove an model and all its components etc */
void glassFreeModel(GlassModel *model)
{
  GlassTexture *tex_ptr = model->textures;
  GlassMaterial *mat_ptr = model->materials;
  GlassSurface *surface_ptr = model->surfaces;
  GlassVariable *var_ptr = model->variables;
  GlassActivePoint *apoint_ptr = model->apoints;
  void *tmp_ptr;

  if(model == NULL)
    return;

  /* Free the textures */
  while(tex_ptr != NULL)
  {
    tmp_ptr = (void *) (tex_ptr->next);
    _glassFreeTexture(tex_ptr);
    tex_ptr = (GlassTexture *) tmp_ptr;
  }

  /* Free the materials */
  while(mat_ptr)
  {
    tmp_ptr = (void *) (mat_ptr->next);
    _glassFreeMaterial(mat_ptr);
    mat_ptr = (GlassMaterial *) tmp_ptr;
  }

  /* Free the surfaces */
  while(surface_ptr)
  {
    tmp_ptr = (void *) (surface_ptr->next);
    _glassFreeSurfaceChild(surface_ptr);
    surface_ptr = (GlassSurface *) tmp_ptr;
  }

  /* Free the variables */
  while(var_ptr)
  {
    tmp_ptr = (void *) (var_ptr->next);
    _glassFreeVariable(var_ptr);
    var_ptr = (GlassVariable *) tmp_ptr;
  }

  /* Free the active points */
  while(apoint_ptr)
  {
    tmp_ptr = (void *) (apoint_ptr->next);
    _glassFreeActivePoint(apoint_ptr);
    apoint_ptr = (GlassActivePoint *) tmp_ptr;
  }

  /* Free the nodes */
  _glassFreeNoder(model->children);

  /* Finally, free the model */
  _glassFreeModel(model);
}

/* Remove a library and all its models */
void glassFreeLibrary(GlassLibrary *library)
{
  GlassLibraryEntry *entry;

  if(!library)
    return;

  entry = library->first;
  while(entry)
  {
    glassFreeModel(entry->model);
    entry->model = NULL;
    entry = entry->next;
  }

  _glassFreeLibrary(library);
}
