/* 
 * glass - the openGL Articulated Skeleton System
 *
 * Copyright (C) 2001-2003  Robert Ancell <bob27@users.sourceforge.net>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

/* glass_load.c
   This file loads a GLASS model file */

/* TODO: Warn about compoenents without names -- apply this to libglassedit too */

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <ctype.h>

#include "global.h"
#include "glass_load.h"
#include "glass_apoint.h"
#include "glass_node.h"
#include "glass_material.h"
#include "glass_model.h"
#include "glass_texture.h"
#include "glass_surface.h"
#include "glass_transform.h"
#include "glass_triangle.h"
#include "glass_variable.h"

/* Stopgap measures as libglassedit uses GLib, but libglass doesn't */

static void *g_malloc0(size_t len)
{
  void *d;

  d = (void *)malloc(len);
  memset(d, 0, len);

  return d;
}

#define g_free(x) free(x)

typedef struct _GList
{
  void *data;
  struct _GList *prev, *next;
} GList;

static GList *g_list_append(GList *list, void *data)
{
  GList *link;

  if(!list)
  {
    link = (GList *)malloc(sizeof(GList));
    link->prev = NULL;
    link->next = NULL;
    link->data = (void *)data;
    return link;
  }

  link = list; 
  while(link->next)
    link = link->next;

  link->next = (GList *)malloc(sizeof(GList));
  link->next->prev = link;
  link->next->next = NULL;
  link->next->data = (void *)data;

  return list;
}

static void g_list_free(GList *list)
{
  GList *t;

  while(list)
  {
    t = list->next;
    free(list);
    list = t;
  }
}

typedef struct
{
  char *str;
} GString;

static void g_string_assign(GString *string, const char *str)
{
  if(string->str)
    free(string->str);

  if(str)
    string->str = strdup(str);
  else
    string->str = NULL;
}

static GString *g_string_new(const char *str)
{
  GString *string;

  string = (GString *)malloc(sizeof(GString));
  string->str = NULL;
  g_string_assign(string, str);

  return string;
}

static void g_string_free(GString *string, boolean free_alloc)
{
  if(!string)
    return;

  if(free_alloc && string->str)
    free(string->str);

  free(string);
}

/* Temporary data to use when loading nodes */
typedef struct
{
  /* Names of components to link to */
  GString *surface_name, *apoint_name, *variable_names[4];

  /* Flag for root nodes */
  int is_root;

  /* Children of nodes */
  int next_index, nchildren, children[MAX_NODE_CHILDREN];
} NodeTempData;

/* Temporary data to use when loading node structure tree */
typedef struct _StructureTempData
{
  /* Children of nodes */
  int index, nchildren, children[MAX_NODE_CHILDREN];
  struct _StructureTempData *next;
} StructureTempData;

static StructureTempData *structure_data = NULL;

static void strip_space(FILE *file);
static void strip_line(FILE *file);
static char *get_raw_line(FILE *file);
static char *get_word(FILE *file, boolean *eol);
static GLfloat get_float(FILE *file, boolean *success, boolean *eol);
static int get_int(FILE *file, boolean *success, boolean *eol);
int get_float_array(FILE *file, GLfloat *array, unsigned int max_entries);
int get_int_array(FILE *file, int *array, unsigned int max_entries);
static void parse_file(GlassLibrary *library, GlassModel *model, FILE *file);
static int find_keyword(const char *keyword_list[], const char *keyword);
static boolean parse_component(GlassModel *model, FILE *file);
static boolean parse_texture(GlassModel *model, FILE *file);
static boolean parse_material(GlassModel *model, FILE *file);
static boolean parse_surface(GlassModel *model, FILE *file);
static boolean parse_triangle(GlassSurface *surface, FILE *file);
static boolean parse_variable(GlassModel *model, FILE *file);
static boolean parse_apoint(GlassModel *model, FILE *file);
static boolean parse_node(GlassModel *model, FILE *file);
static boolean parse_structure(GlassModel *model, FILE *file);
static int next_structure_index(char *line, int column, int *index, boolean *has_siblings);
static int get_structure_index(char *line, int column, boolean *has_siblings);
static void get_children(GlassModel *model, char *line, int column, int index);
static NodeTempData *allocate_node_temp_data(void);
static void free_node_temp_data(NodeTempData *data);
static void link_material(GlassModel *model, GlassMaterial *material);
static void sort_triangles(GlassSurface *surface);
static void link_triangle(GlassModel *model, GlassTriangle *triangle);
static GlassNode *find_node(int index);
static void convert_structure_tree(GlassModel *model);
static void link_nodes(GlassModel *model, GlassNode *node);
static void test_node_loops(GlassModel *model);
static boolean test_loop_r(GlassNode *node);

/* Words that are used in the file format */
enum {KW_NAME = 0, KW_VERSION, KW_BEGIN};
static const char *keywords[] = {"name", "version", "begin", NULL};

enum {KW_COM_TEXTURE = 0, KW_COM_MATERIAL, KW_COM_SURFACE, KW_COM_SURFACE2, KW_COM_VARIABLE,
      KW_COM_APOINT, KW_COM_APOINT2, KW_COM_APOINT3, KW_COM_NODE, KW_COM_NODE2, KW_COM_STRUCTURE};
static const char *component_keywords[] = {"texture", "material", "surface", "triangle_list",
					   "variable", "active point", "active_point",
					   "apoint", "component", "node", "structure", NULL};

enum {KW_TEX_NAME = 0, KW_TEX_IMAGlass, KW_TEX_END};
static const char *texture_keywords[] = {"name", "image", "end", NULL};

enum {KW_MAT_NAME = 0, KW_MAT_TEXTURE, KW_MAT_AMBIENT, KW_MAT_DIFFUSE, KW_MAT_SPECULAR,
      KW_MAT_EMISSION, KW_MAT_SHININESS, KW_MAT_END};
static const char *material_keywords[] = {"name", "texture", "ambient", "diffuse",
					  "specular", "emission", "shininess", "end", NULL};

enum {KW_SURF_NAME = 0, KW_SURF_TRIANGLE, KW_SURF_END};
static const char *surface_keywords[] = {"name", "{", "end", NULL};

enum {KW_TRI_MATERIAL = 0, KW_TRI_POS, KW_TRI_COL, KW_TRI_ALPHA, KW_TRI_NOR,
      KW_TRI_TEX, KW_TRI_END};
static const char *triangle_keywords[] = {"material", "pos", "col", "alpha", "nor",
					  "tex", "}", NULL};

enum {KW_VAR_NAME = 0, KW_VAR_MIN, KW_VAR_MAX, KW_VAR_VALUE, KW_VAR_END};
static const char *variable_keywords[] = {"name", "min", "max", "value", "end", NULL};

enum {KW_AP_NAME = 0, KW_AP_DIR, KW_AP_DIR2, KW_AP_UPDATE, KW_AP_END};
static const char *apoint_keywords[] = {"name", "initial_dir", "dir", "update", "end", NULL};

enum {KW_NODE_INDEX = 0, KW_NODE_CHILDREN, KW_NODE_CHILDREN2, KW_NODE_CHILDREN3,
      KW_NODE_SURFACE, KW_NODE_SURFACE2, KW_NODE_APOINT, KW_NODE_APOINT2, KW_NODE_APOINT3,
      KW_NODE_TRANSFORM, KW_NODE_END};
static const char *node_keywords[] = {"index", "child", "children", "child_index", "surface",
				      "list", "active point", "active_point", "apoint",
				      "transform", "end", NULL};

static const char *transform_types[] = {"NONE", "TRANSLATE", "ROTATE", "SCALE", "TRANSLATE_X",
					"TRANSLATE_Y", "TRANSLATE_Z", "ROTATE_X", "ROTATE_Y",
					"ROTATE_Z", "SCALE_X", "SCALE_Y", "SCALE_Z",
					"SCALE_EVEN", NULL};


static const char *update_types[] = {"NONE", "CONTINUOUS", "DEMAND", NULL};

/* Line counter */
static int line_num;

/* Filename of model that's being worked on */
static GString *model_fname;

/* Partially loaded nodes */
GList *node_list = NULL;

/* Load the model file 'fname', use a library if one is given */
GlassModel *_glassLoadModel(GlassLibrary *library, const char *fname)
{
  GlassModel *model;
  FILE *file;

  file = fopen(fname, "r");
  if(!file)
  {
    fprintf(stderr, "Glass Error: Unable to open model file \"%s\"\n", fname);
    return NULL;
  }

  model = _glassModelNew("");
  if(!model)
    return NULL;
  if(library)
    _glassLibraryAddModel(library, model);

  model_fname = g_string_new(fname);

  /* Reset the line counter */
  line_num = 1;

  structure_data = NULL;
  node_list = NULL;

  /* Load the components from the file */
  parse_file(library, model, file);

  g_string_free(model_fname, TRUE);

  fclose(file);

  /* Build some arrays that link to the variables and the surfaces for fast access */
  _glassModelBuildArrays(model);

  /* Convert the surfaces into openGL display lists if possible */
  _glassModelMakeDisplayLists(model);

  /* Work out which nodes are required for active point calculations */
  _glassNodeTestActivePointr(model->children);

  /* Finally! The glass model in all it's glory -- hopefully */
  return model;
}

/* Strip space and comments from a file */
static void strip_space(FILE *file)
{
  int c;

  /* Pull off characters until a non-whitespace one is reached */
  do {
    c = fgetc(file);
    switch(c)
    {
    case '\n':
      line_num++;
      break;
    case '#':
      /* Comment, strip the rest of this line off */
      strip_line(file);
      break;
    default:
      break;
    }
  } while(isspace(c) || c == '#');

  /* File finished */
  if(c == EOF)
    return;

  /* Push the last character back onto the stream */
  ungetc(c, file);
}

/* Strip characters until the next line */
static void strip_line(FILE *file)
{
  int c;

  do {
    c = fgetc(file);
  } while(c != EOF && c != '\n');

  if(c == '\n')
    line_num++;
}

/* Get a raw line from the file (i.e. don't remove whitespace etc). */
static char *get_raw_line(FILE *file)
{
  static char line[MAX_LINE_LENGTH];
  int c, i;

  /* Check if there is a line to get */
  c = fgetc(file);
  if(c == '\n')
  {
    line_num++;
    c = fgetc(file);
  }
  if(c == EOF)
    return NULL;
  line[0] = c;

  /* Get the rest of the line */
  for(i = 1; i < MAX_LINE_LENGTH; i++)
  {
    c = fgetc(file);
    if(c == EOF)
      break;

    if(c == '\n')
    {
      ungetc(c, file);
      break;
    }

    line[i] = c;
  }
  line[i] = '\0';

  return line;
}

/* Get the next word from a file. The string is contained internally and needs
   to be copied if required.  Write TRUE into eol if this word is on the end of a line */
static char *get_word(FILE *file, boolean *eol)
{
  int c, word_index = 0;
  boolean in_quotes = FALSE;
  static char word[MAX_WORD_LENGTH];

  /* Strip whitespace and comments off the file */
  strip_space(file);

  /* Check if is contained in quotes (") */
  c = fgetc(file);
  if(c == EOF)
    return NULL;
  if(c == '\"')
    in_quotes = TRUE;
  else
    word[word_index++] = c;

  /* Get the rest of the word */
  do {
    c = fgetc(file);

    /* Quoted words end with a quote, non quoted end with whitespace */
    if(in_quotes)
    {
      if(c == '\"')
	break;
    }
    else if(isspace(c))
      break;
	
    word[word_index++] = c;
  } while(word_index < MAX_WORD_LENGTH && c != EOF);

  /* Push newlines back onto the stream so the line number info is
     correct for _this_ word */
  if(c == '\n')
  {
    ungetc(c, file);
    if(eol)
      *eol = TRUE;
  }
  else
  {
    if(eol)
      *eol = FALSE;
  }

  /* Terminate string */
  word[word_index] = '\0';

  return word;
}

/* Get the next word from a file and convert to a floating point number.
   Write the success into the pointer success. Write TRUE into eol if
   this number is on the end of a line */
static GLfloat get_float(FILE *file, boolean *success, boolean *eol)
{
  const char *word;

  word = get_word(file, eol);
  if(!word)
  {
    if(success)
      *success = FALSE;
    return 0.0;
  }

  if(success)
    *success = TRUE;

  return atof(word);
}

/* Get the next word from a file and convert to an integer number.
   Write the success into the pointer success. Write TRUE into eol if
   this number is on the end of a line */
static int get_int(FILE *file, boolean *success, boolean *eol)
{
  const char *word;

  word = get_word(file, eol);
  if(!word)
  {
    if(success)
      *success = FALSE;
    return 0.0;
  }

  if(success)
    *success = TRUE;

  return atoi(word);
}

/* Get at most max_entries floating point numbers and store them in array.
   The numbers must be contained in one line. Return the number of entries
   loaded. */
int get_float_array(FILE *file, GLfloat *array, unsigned int max_entries)
{
  int entry;
  boolean eol, success;

  eol = FALSE;
  entry = 0;
  do {
    array[entry++] = get_float(file, &success, &eol);
    if(!success)
      return entry - 1;
  } while(!eol && entry < max_entries);

  return entry;
}

/* Get at most max_entries integer numbers and store them in array.
   The numbers must be contained in one line. Return the number of entries
   loaded. */
int get_int_array(FILE *file, int *array, unsigned int max_entries)
{
  int entry;
  boolean eol, success;

  eol = FALSE;
  entry = 0;
  do {
    array[entry++] = get_int(file, &success, &eol);
    if(!success)
      return entry - 1;
  } while(!eol && entry < max_entries);

  return entry;
}

/* Load the components from the file */
static void parse_file(GlassLibrary *library, GlassModel *model, FILE *file)
{
  char *word;
  int keyword_index;
  GList *link;
  GlassMaterial *material;
  GlassSurface *surface;
  GlassTriangle *triangle;
  GlassNode *root_node;

  /* Load the model paramters and components */
  while((word = get_word(file, NULL)))
  {
    keyword_index = find_keyword(keywords, word);

    switch(keyword_index)
    {
    case -1:
      fprintf(stderr, "Glass Error: Unknown keyword \"%s\" in model file \"%s\", line %d\n",
	      word, model_fname->str, line_num);
      break;
    case KW_NAME:
      word = get_word(file, NULL);
      _glassModelSetName(model, word);
      break;
    case KW_VERSION:
      word = get_word(file, NULL);
      /* TODO: What to do here? Compare with the version that glassedit is designed for? */
      break;
    case KW_BEGIN:
      parse_component(model, file);
      break;
    default:
      fprintf(stderr, "Glass Error: Keyword \"%s\" not handled (%s::%d)\n", word, __FILE__, __LINE__);
      break;
    }
  }

  /* Convert a node structure tree into child arrays (if one supplied) */
  if(structure_data)
    convert_structure_tree(model);

  /* Link up the components (must wait until they are all loaded) */
  material = model->materials;
  while(material)
  {
    link_material(model, material);
    material = material->next;
  }
  surface = model->surfaces;
  while(surface)
  {
    /* Convert material names into materials */
    triangle = surface->triangles;
    while(triangle)
    {
      link_triangle(model, triangle);
      triangle = triangle->next;
    }

    /* Sort the transparent triangles from the list */
    sort_triangles(surface);

    surface = surface->next;
  }

  root_node = find_node(0);
  link = node_list;
  while(link)
  {
    link_nodes(model, link->data);
    link = link->next;
  }
  /* If didn't put in any nodes, 0 is the default root node */
  if(!model->children)
    _glassModelAddChild(model, root_node);

  /* Test for loops in the structure */
  test_node_loops(model);
}

/* Return the index of a keyword from a NULL terminated list of keywords (keyword_list).
   Return -1 if not in the list. */
static int find_keyword(const char *keyword_list[], const char *keyword)
{
  int index;

  if(!keyword_list)
    return -1;

  /* Find the word in the list */
  for(index = 0; keyword_list[index] && strcmp(keyword, keyword_list[index]); index++);
  if(keyword_list[index])
    return index;

  return -1;
}

/* Parse a component (comes after a begin keyword). Return TRUE if successfully
   completed. */
static boolean parse_component(GlassModel *model, FILE *file)
{
  const char *word;
  int index;

  word = get_word(file, NULL);
  if(!word)
    return FALSE;

  index = find_keyword(component_keywords, word);
  switch(index)
  {
  case -1:
    fprintf(stderr, "Glass Error: Unknown component type \"%s\" in model file \"%s\", line %d\n",
	    word, model_fname->str, line_num);
    break;
  case KW_COM_TEXTURE:
    return parse_texture(model, file);
    break;
  case KW_COM_MATERIAL:
    return parse_material(model, file);
    break;
  case KW_COM_SURFACE:
  case KW_COM_SURFACE2:
    return parse_surface(model, file);
    break;
  case KW_COM_VARIABLE:
    return parse_variable(model, file);
    break;
  case KW_COM_APOINT:
  case KW_COM_APOINT2:
  case KW_COM_APOINT3:
    return parse_apoint(model, file);
    break;
  case KW_COM_NODE:
  case KW_COM_NODE2:
    return parse_node(model, file);
    break;
  case KW_COM_STRUCTURE:
    return parse_structure(model, file);
    break;
  default:
    fprintf(stderr, "Glass Error: Component keyword \"%s\" not handled (%s::%d)\n",
	    word, __FILE__, __LINE__);
    break;
  }

  return FALSE;
}

/* Parse a texture */
static boolean parse_texture(GlassModel *model, FILE *file)
{
  GlassTexture *texture;
  const char *word;
  char *dir;
  int i, l, index;

  texture = _glassTextureNew();
  if(!texture)
    return FALSE;
  _glassModelAddTexture(model, texture);

  /* Get the parameters for the texture */
  while((word = get_word(file, NULL)))
  {
    if(!word)
      return FALSE;

    index = find_keyword(texture_keywords, word);
    switch(index)
    {
    case -1:
      fprintf(stderr, "Glass Error: Unknown texture parameter \"%s\" in model file \"%s\", line %d\n",
	      word, model_fname->str, line_num);
      break;
    case KW_TEX_NAME:
      word = get_word(file, NULL);
      _glassTextureSetName(texture, word);
      break;
    case KW_TEX_IMAGlass:
      word = get_word(file, NULL);
      /* Get the directory that the glass model is in */
      /* FIXME: Improve this and apply in libglassedit */
      dir = strdup(model_fname->str);
      l = strlen(dir);
      for(i = l-2; i >= 0; i--)
      {
	if(dir[i] == '/') /* NOTE change to '\' for win32 */
	{
	  dir[i+1] = '\0';
	  break;
	}
      }
      if(i < 0)
	_glassTextureLoadRGB(texture, word);
      else
      {
#define MAX_NAME_LENGTH 1024
	char texture_file[MAX_NAME_LENGTH];

	snprintf(texture_file, MAX_NAME_LENGTH, "%s%s", dir, word);
	_glassTextureLoadRGB(texture, texture_file);
      }
      break;
    case KW_TEX_END:
      return TRUE;
      break;
    default:
      fprintf(stderr, "Glass Error: Texture keyword \"%s\" not handled (%s::%d)\n",
	      word, __FILE__, __LINE__);
      break;
    }
  }

  return FALSE;
}

/* Parse a material */
static boolean parse_material(GlassModel *model, FILE *file)
{
  GlassMaterial *material;
  const char *word;
  int index, nentries;
  GLfloat value, array[4];
  boolean success;

  material = _glassMaterialNew();
  if(!material)
    return FALSE;
  _glassModelAddMaterial(model, material);

  /* Store the name of the texture that will be linked to */
  material->ad = (void *)g_string_new("");

  /* Get the parameters for the material */
  while((word = get_word(file, NULL)))
  {
    if(!word)
      return FALSE;

    index = find_keyword(material_keywords, word);
    switch(index)
    {
    case -1:
      fprintf(stderr, "Glass Error: Unknown material parameter \"%s\" in model file \"%s\", line %d\n",
	      word, model_fname->str, line_num);
      break;
    case KW_MAT_NAME:
      word = get_word(file, NULL);
      _glassMaterialSetName(material, word);
      break;
    case KW_MAT_TEXTURE:
      word = get_word(file, NULL);
      g_string_assign((GString *)material->ad, word);
      break;
    case KW_MAT_AMBIENT:
      /* Default alpha of 1.0 */
      array[3] = 1.0;
      nentries = get_float_array(file, array, 4);
      if(nentries != 3 && nentries != 4)
	fprintf(stderr, "Glass Error: Incorrect number of values (%d) for ambient colour of texture in "
		"model file \"%s\", line %d. Ambient colour ignored\n",
		nentries, model_fname->str, line_num);
      else
	_glassMaterialSetAmbientv(material, array);
      break;
    case KW_MAT_DIFFUSE:
      /* Default alpha of 1.0 */
      array[3] = 1.0;
      nentries = get_float_array(file, array, 4);
      if(nentries != 3 && nentries != 4)
	fprintf(stderr, "Glass Error: Incorrect number of values (%d) for diffuse colour of texture in "
		"model file \"%s\", line %d. Diffuse colour ignored\n",
		nentries, model_fname->str, line_num);
      else
	_glassMaterialSetDiffusev(material, array);
      break;
    case KW_MAT_SPECULAR:
      /* Default alpha of 1.0 */
      array[3] = 1.0;
      nentries = get_float_array(file, array, 4);
      if(nentries != 3 && nentries != 4)
	fprintf(stderr, "Glass Error: Incorrect number of values (%d) for specular colour of texture in "
		"model file \"%s\", line %d. Specular colour ignored\n",
		nentries, model_fname->str, line_num);
      else
	_glassMaterialSetSpecularv(material, array);
      break;
    case KW_MAT_EMISSION:
      /* Default alpha of 1.0 */
      array[3] = 1.0;
      nentries = get_float_array(file, array, 4);
      if(nentries != 3 && nentries != 4)
	fprintf(stderr, "Glass Error: Incorrect number of values (%d) for emission colour of texture in "
		"model file \"%s\", line %d. Emission colour ignored\n",
		nentries, model_fname->str, line_num);
      else
	_glassMaterialSetEmissionv(material, array);
      break;
    case KW_MAT_SHININESS:
      value = get_float(file, &success, NULL);
      if(success)
	_glassMaterialSetShininess(material, value);
      else
	fprintf(stderr, "Glass Error: Shininess paramter in texture component without a value in "
		"model file \"%s\", line %d\n", model_fname->str, line_num);
      break;
    case KW_MAT_END:
      return TRUE;
      break;
    default:
      fprintf(stderr, "Glass Error: Material keyword \"%s\" not handled (%s::%d)\n",
	      word, __FILE__, __LINE__);
      break;
    }
  }

  return FALSE;
}

/* Parse a surface */
static boolean parse_surface(GlassModel *model, FILE *file)
{
  GlassSurface *surface;
  const char *word;
  int index;

  surface = _glassSurfaceNew("");
  if(!surface)
    return FALSE;
  _glassModelAddSurface(model, surface);

  /* Get the parameters for the surface */
  while((word = get_word(file, NULL)))
  {
    index = find_keyword(surface_keywords, word);
    switch(index)
    {
    case -1:
      fprintf(stderr, "Glass Error: Unknown surface parameter \"%s\" in model file \"%s\", line %d\n",
	      word, model_fname->str, line_num);
      break;
    case KW_SURF_NAME:
      word = get_word(file, NULL);
      _glassSurfaceSetName(surface, word);
      break;
    case KW_SURF_TRIANGLE:
      parse_triangle(surface, file);
      break;
    case KW_SURF_END:
      return TRUE;
      break;
    default:
      fprintf(stderr, "Glass Error: Surface keyword \"%s\" not handled (%s::%d)\n",
	      word, __FILE__, __LINE__);
      break;
    }
  }

  return FALSE;
}

/* Parse a triangle */
/* FIXME: Convert into alpha triangles */
static boolean parse_triangle(GlassSurface *surface, FILE *file)
{
  GlassTriangle *triangle;
  GString *material_name;
  const char *word;
  int i, index, vertex;
  GLfloat value, array[4];
  boolean success;

  triangle = _glassTriangleNew(surface);
  if(!triangle)
    return FALSE;

  /* Store the name of the material that will be linked to */
  material_name = g_string_new("");
  triangle->ad = (void *)material_name;

  /* Get the parameters for the triangle */
  while((word = get_word(file, NULL)))
  {
    index = find_keyword(triangle_keywords, word);

    switch(index)
    {
    case -1:
      fprintf(stderr, "Glass Error: Unknown triangle parameter \"%s\" in model file \"%s\", line %d\n",
	      word, model_fname->str, line_num);
      break;
    case KW_TRI_MATERIAL:
      word = get_word(file, NULL);
      g_string_assign(material_name, word);
      break;
    case KW_TRI_POS:
    case KW_TRI_COL:
    case KW_TRI_ALPHA:
    case KW_TRI_NOR:
    case KW_TRI_TEX:
      /* Get which vertex this is */
      vertex = get_int(file, &success, NULL);
      if(!success)
      {
	fprintf(stderr, "Glass Error: Triangle %s parameter with no vertex index "
		"in model file \"%s\", line %d\n",
		triangle_keywords[index], model_fname->str, line_num);
	break;
      }
      else if(vertex < 0 || vertex > 2)
      {
	fprintf(stderr, "Glass Error: Invalid vertex (%d) for %s triangle parameter "
		"in model file \"%s\", line %d\n",
		vertex,  triangle_keywords[index],
		model_fname->str, line_num);
	break;
      }

      /* Get the data */
      switch(index)
      {
      case KW_TRI_POS:
	array[0] = 0.0;
	array[1] = 0.0;
	array[2] = 0.0;
	get_float_array(file, array, 3);
	for(i = 0; i < 3; i++)
	  triangle->pos[vertex][i] = array[i];
	break;
      case KW_TRI_COL:
	array[0] = 0.0;
	array[1] = 0.0;
	array[2] = 0.0;
	array[3] = 1.0;
	get_float_array(file, array, 4);
	for(i = 0; i < 4; i++)
	  triangle->col[vertex][i] = array[i];
	break;
      case KW_TRI_ALPHA:
	value = get_float(file, &success, NULL);
	if(!success)
	{
	  fprintf(stderr, "Glass Error: Triangle alpha parameter with no value "
		  "in model file \"%s\", line %d\n",
		  model_fname->str, line_num);
	  break;
	}
	triangle->col[vertex][3] = value;
	break;
      case KW_TRI_NOR:
	array[0] = 0.0;
	array[1] = 0.0;
	array[2] = 1.0;
	get_float_array(file, array, 3);
	for(i = 0; i < 3; i++)
	  triangle->nor[vertex][i] = array[i];
	break;
      case KW_TRI_TEX:
	array[0] = 0.0;
	array[1] = 0.0;
	get_float_array(file, array, 2);
	for(i = 0; i < 2; i++)
	  triangle->tex[vertex][i] = array[i];
	break;
      default:
	break;
      }
      break;
    case KW_TRI_END:
      return TRUE;
      break;
    default:
      fprintf(stderr, "Glass Error: Triangle keyword \"%s\" not handled (%s::%d)\n",
	      word, __FILE__, __LINE__);
      break;
    }
  }

  return FALSE;
}

/* Parse a variable */
static boolean parse_variable(GlassModel *model, FILE *file)
{
  GlassVariable *variable;
  const char *word;
  GLfloat value;
  int index;
  boolean success;

  variable = _glassVariableNew("", 0.0, 1.0, 0.0);
  if(!variable)
    return FALSE;
  _glassModelAddVariable(model, variable);

  /* Get the parameters for the variable */
  while((word = get_word(file, NULL)))
  {
    if(!word)
      return FALSE;

    index = find_keyword(variable_keywords, word);
    switch(index)
    {
    case -1:
      fprintf(stderr, "Glass Error: Unknown variable parameter \"%s\" in model file \"%s\", line %d\n",
	      word, model_fname->str, line_num);
      break;
    case KW_VAR_NAME:
      word = get_word(file, NULL);
      _glassVariableSetName(variable, word);
      break;
    case KW_VAR_MIN:
    case KW_VAR_MAX:
    case KW_VAR_VALUE:
      /* Get the value */
      value = get_float(file, &success, NULL);
      if(!success)
      {
	fprintf(stderr, "Glass Error: Variable parameter %s with no value "
		"in model file \"%s\", line %d\n",
		variable_keywords[index], model_fname->str, line_num);
	break;
      }

      /* Apply the value */
      switch(index)
      {
      case KW_VAR_MIN:
	variable->min = value;
	break;
      case KW_VAR_MAX:
	variable->max = value;
	break;
      case KW_VAR_VALUE:
	variable->value = value;
	break;
      default:
	break;
      }
      break;
    case KW_VAR_END:
      return TRUE;
      break;
    default:
      fprintf(stderr, "Glass Error: Variable keyword \"%s\" not handled (%s::%d)\n",
	      word, __FILE__, __LINE__);
      break;
    }
  }

  return FALSE;
}

/* Parse a active point */
static boolean parse_apoint(GlassModel *model, FILE *file)
{
  GlassActivePoint *apoint;
  const char *word;
  int index, update_index;
  GLfloat array[3];

  apoint = _glassActivePointNew();
  if(!apoint)
    return FALSE;
  _glassModelAddActivePoint(model, apoint);

  /* Get the parameters for the triangle */
  while((word = get_word(file, NULL)))
  {
    if(!word)
      return FALSE;

    index = find_keyword(apoint_keywords, word);
    switch(index)
    {
    case -1:
      fprintf(stderr, "Glass Error: Unknown active point parameter \"%s\" in model file \"%s\", line %d\n",
	      word, model_fname->str, line_num);
      break;
    case KW_AP_NAME:
      word = get_word(file, NULL);
      _glassActivePointSetName(apoint, word);
      break;
    case KW_AP_DIR:
    case KW_AP_DIR2:
      array[0] = 0.0;
      array[1] = 0.0;
      array[2] = 0.0;
      array[3] = 1.0;
      get_float_array(file, array, 3);
      _glassActivePointSetInitDirv(apoint, array);
      break;
    case KW_AP_UPDATE:
      word = get_word(file, NULL);
      update_index = find_keyword(update_types, word);
      switch(update_index)
      {
      case -1:
	fprintf(stderr, "Glass Error: Unknown active point update parameter \"%s\" "
		"in model file \"%s\", line %d\n",
		word, model_fname->str, line_num);
	break;
      case GLASS_UPDATE_NONE:
      case GLASS_UPDATE_CONTINUOUS:
      case GLASS_UPDATE_DEMAND:
	apoint->update = update_index;
      default:
	fprintf(stderr, "Glass Error: Update type \"%s\" not handled (%s::%d)\n", word, __FILE__, __LINE__);
	break;
      }
      break;
    case KW_AP_END:
      return TRUE;
      break;
    default:
      fprintf(stderr, "Glass Error: Active point keyword \"%s\" not handled (%s::%d)\n",
	      word, __FILE__, __LINE__);
      break;
    }
  }

  return FALSE;
}

/* Parse a node */
static boolean parse_node(GlassModel *model, FILE *file)
{
  GlassNode *node;
  GlassTransform *transform;
  NodeTempData *td;
  const char *word;
  int index, transform_type, param, nparams;
  boolean success;

  node = _glassNodeNew();
  if(!node)
    return FALSE;
  node_list = g_list_append(node_list, node);

  /* Store the names of surface, active points, and the child
     indicies of this node */
  td = allocate_node_temp_data();
  node->ad = (void *)td;

  /* Get the parameters for the node */
  while((word = get_word(file, NULL)))
  {
    if(!word)
      return FALSE;

    index = find_keyword(node_keywords, word);
    switch(index)
    {
    case -1:
      fprintf(stderr, "Glass Error: Unknown node parameter \"%s\" in model file \"%s\", line %d\n",
	      word, model_fname->str, line_num);
      break;
    case KW_NODE_INDEX:
      node->index = get_int(file, &success, NULL);
      break;
    case KW_NODE_CHILDREN:
    case KW_NODE_CHILDREN2:
    case KW_NODE_CHILDREN3:
      td->nchildren = get_int_array(file, td->children, MAX_NODE_CHILDREN);
      break;
    case KW_NODE_SURFACE:
    case KW_NODE_SURFACE2:
      word = get_word(file, NULL);
      g_string_assign(td->surface_name, word);
      break;
    case KW_NODE_APOINT:
    case KW_NODE_APOINT2:
    case KW_NODE_APOINT3:
      word = get_word(file, NULL);
      g_string_assign(td->apoint_name, word);
      break;
    case KW_NODE_TRANSFORM:
      word = get_word(file, NULL);
      transform_type = find_keyword(transform_types, word);
      transform = _glassTransformNew(transform_type);
      _glassNodeSetTransform(node, transform);
      if(transform_type == -1)
      {
	fprintf(stderr, "Glass Error: Unknown node transform \"%s\" in model file \"%s\", line %d\n",
		word, model_fname->str, line_num);
	break;
      }

      /* Get the number of paramters expected */
      nparams = _glassTransformGetnParams(transform_type);

      /* Try to get this many parameters */
      for(param = 0; param < nparams; param++)
      {
	word = get_word(file, NULL);
	if(!word)
	{
	  fprintf(stderr, "Glass Error: Incorrect number of transform parameters found (%d/%d) "
		  "in model file \"%s\", line %d\n",
		  param - 1, nparams, model_fname->str, line_num);
	  break;
	}

	/* Check if is a variable or a constant */
	if(tolower(word[0]) == 'v')
	{
	  if(strncmp(word + 1, "::", 2))
	  {
	    fprintf(stderr, "Glass Error: Invalid node variable parameter \"%s\" "
		    "in model file \"%s\", line %d\n",
		    word, model_fname->str, line_num);
	    continue;
	  }
	  g_string_assign(td->variable_names[param], word + 3);
	}
	else
	  _glassTransformSetConstant(transform, param, atof(word));
      }
      break;
    case KW_NODE_END:
      return TRUE;
      break;
    default:
      fprintf(stderr, "Glass Error: Node keyword \"%s\" not handled (%s::%d)\n", word, __FILE__, __LINE__);
      break;
    }
  }

  return FALSE;
}

/* Parse the structure data */
static boolean parse_structure(GlassModel *model, FILE *file)
{
  GList *lines, *entry;
  GString *string;
  char *line, *next_line;
  int init_line_num, line_num_offset, index, column;
  boolean has_siblings;

  init_line_num = line_num;

  /* Copy the structure lines */
  lines = NULL;
  while((line = get_raw_line(file)))
  {
    if(!strcmp(line, "end"))
      break;

    string = g_string_new(line);
    lines = g_list_append(lines, string);
  }

  if(!line)
    fprintf(stderr, "Glass Error: Node structure finished without 'end' in model file \"%s\", line %d\n",
	    model_fname->str, line_num);

  if(structure_data)
    fprintf(stderr, "Glass Error: Duplicate node structure found in model file \"%s\", line %d, "
	    "duplicate ignored\n", model_fname->str, line_num);
  else
  {
    /* Convert the structure lines */
    line_num_offset = 0;
    entry = lines;

    /* Get the root children */
    if(entry)
      get_children(model, ((GString *)entry->data)->str, 0, -1);

    while(entry)
    {
      /* Get the lines to work on */
      line = ((GString *)entry->data)->str;
      if(entry->next)
	next_line = ((GString *)entry->next->data)->str;
      else
	next_line = NULL;

      /* Go through each entry on this line */
      for(column = 0; ; column++)
      {
	/* Get the next node index */
	column = next_structure_index(line, column, &index, &has_siblings);
	if(column == -1)
	  break;

	if(index == -1)
	  fprintf(stderr, "Glass Error: Invalid node index in model file \"%s\", line %d\n",
		  model_fname->str, init_line_num + line_num_offset);
	else
	  get_children(model, next_line, column, index);
      }

      line_num_offset++;
      entry = entry->next;
    }
  }

  /* Free the structure lines */
  entry = lines;
  while(entry)
  {
    g_string_free((GString *)entry->data, TRUE);
    entry = entry->next;
  }
  g_list_free(lines);

  return FALSE;
}

/* Look along a line for the next structure index starting at the given column. Return
   the column this index is in or -1 if there are no available indicies. Write the index
   into 'index'. If this index has a sibling write this into 'has_siblings' */
static int next_structure_index(char *line, int column, int *index, boolean *has_siblings)
{
  int ncolumns;

  ncolumns = (strlen(line) + 1) / 4;
  if(column >= ncolumns)
    return -1;

  for(; column < ncolumns; column++)
  {
    *index = get_structure_index(line, column, has_siblings);
    if(*index >= 0)
      return column;
    if(*index == -2)
      break;
  }

  return -1;
}

/* Get the node index at the given column of the line. Return -1 if no index at this
   column. Return -2 if run out of line. If the index has a sibling write this into
   'has_siblings' */
static int get_structure_index(char *line, int column, boolean *has_siblings)
{
  int index = -1, ncolumns;
  char index_string[4], *part_line;

  ncolumns = (strlen(line) + 1) / 4;
  if(column >= ncolumns)
    return -2;

  /* Convert the string to an index at this column */
  part_line = line + column * 4;
  if(part_line[0] != ' ' && part_line[0] != '-')
  {
    index_string[0] = part_line[0];
    index_string[1] = part_line[1];
    index_string[2] = part_line[2];
    index_string[3] = '\0';
    if(!isdigit(index_string[0]) || !isdigit(index_string[1]) || !isdigit(index_string[2]))
    {
      fprintf(stderr, "Glass Error: Invalid node index %c%c%c\n",
	      index_string[0], index_string[1], index_string[2]);
    }
    else
      index = atoi(index_string);
  }

  /* Line may continue on (siblings) */
  if(part_line[3] == '-')
    *has_siblings = TRUE;
  else
    *has_siblings = FALSE;

  return index;
}

/* Get the children from the given line starting at the given column. These are
   the children for the node with index 'index' */
static void get_children(GlassModel *model, char *line, int column, int index)
{
  StructureTempData *td;
  int child;
  boolean has_siblings;

  if(!line)
    return;

  td = (StructureTempData *)malloc(sizeof(StructureTempData));
  memset(td, 0, sizeof(*td));
  td->next = structure_data;
  structure_data = td;

  td->index = index;
  while(TRUE) {
    /* Look for the next index */
    child = get_structure_index(line, column, &has_siblings);

    /* Out of line */
    if(child == -2)
      return;

    /* If a child, store it */
    if(child >= 0)
    {
      if(td->nchildren >= MAX_NODE_CHILDREN)
	return;
      td->children[td->nchildren++] = child;
    }

    /* More children? */
    if(has_siblings)
      column++;
    else
      return;
  }
}

/* Allocate temporary data for nodes */
static NodeTempData *allocate_node_temp_data(void)
{
  NodeTempData *data;

  data = g_malloc0(sizeof(NodeTempData));
  data->surface_name = g_string_new("");
  data->apoint_name = g_string_new("");
  data->variable_names[0] = g_string_new("");
  data->variable_names[1] = g_string_new("");
  data->variable_names[2] = g_string_new("");
  data->variable_names[3] = g_string_new("");
  data->next_index = -1;

  return data;
}

/* Free temporary data for nodes */
static void free_node_temp_data(NodeTempData *data)
{
  g_string_free(data->surface_name, TRUE);
  g_string_free(data->apoint_name, TRUE);
  g_string_free(data->variable_names[0], TRUE);
  g_string_free(data->variable_names[1], TRUE);
  g_string_free(data->variable_names[2], TRUE);
  g_string_free(data->variable_names[3], TRUE);
  g_free(data);
}

/* Convert temporary texture links to actual textures */
static void link_material(GlassModel *model, GlassMaterial *material)
{
  GString *texture_name;
  GlassTexture *texture;

  texture_name = (GString *)material->ad;
  texture = _glassModelFindTexture(model, texture_name->str);
  if(texture)
    _glassMaterialSetTexture(material, texture);
  else
    fprintf(stderr, "Glass Error: Unable to find texture \"%s\"\n", texture_name->str);

  g_string_free(texture_name, TRUE);
  material->ad = NULL;
}

/* Sort the transparent triangles from the opaque ones */
static void sort_triangles(GlassSurface *surface)
{
  GlassTriangle *triangle, *t, *l;

  triangle = surface->triangles;
  l = NULL;
  while(triangle)
  {
    t = triangle->next;

    /* Move into the alpha list */
    if(triangle->is_alpha)
    {
      /* Skip this triangle out */
      if(triangle == surface->last_triangle)
	surface->last_triangle = l;
      if(l)
	l->next = t;
      else
	surface->triangles = t;

      /* Append to the transparent list */
      if(surface->last_alpha)
      {
	surface->last_alpha->next = triangle;
	surface->last_alpha = triangle;
      }
      else
      {
	surface->alpha_triangles = triangle;
	surface->last_alpha = triangle;
      }
    }
    else
      /* Move onto the next triangle */
      l = triangle;

    triangle = t;
  }
}

/* Convert temporary material links to actual materials */
static void link_triangle(GlassModel *model, GlassTriangle *triangle)
{
  GString *material_name;
  GlassMaterial *material;

  /* Link up the material (if has one) */
  material_name = (GString *)triangle->ad;
  if(strcmp(material_name->str, ""))
  {
    material = _glassModelFindMaterial(model, material_name->str);
    if(material)
      _glassTriangleSetMaterial(triangle, material);
    else
      fprintf(stderr, "Glass Error: Unable to find material \"%s\"\n", material_name->str);
  }

  /* Check if the triangle is transparent */
  if(triangle->col[0][3] != 1.0 || triangle->col[1][3] != 1.0 ||
     triangle->col[2][3] != 1.0 || (material && material->is_alpha))
    triangle->is_alpha = TRUE;

  g_string_free(material_name, TRUE);
  triangle->ad = NULL;
}

/* Find a node */
static GlassNode *find_node(int index)
{
  GList *link;
  GlassNode *node;

  link = node_list;
  while(link)
  {
    node = (GlassNode *)link->data;

    if(node->index == index)
      return node;

    link = link->next;
  }

  return NULL;
}

/* Convert the node structure component into child arrays */
/* TODO: Check for duplicate entries */
static void convert_structure_tree(GlassModel *model)
{
  StructureTempData *td, *t;
  NodeTempData *ntd;
  GlassNode *node;
  int i;

  td = structure_data;
  while(td)
  {
    /* The root nodes */
    if(td->index == -1)
    {
      for(i = 0; i < td->nchildren; i++)
      {
	node = find_node(td->children[i]);
	if(node)
	{
	  ntd = (NodeTempData *)node->ad;
	  ntd->is_root = TRUE;
	}
	/* TODO: What if node not found! */
      }
    }
    else
    {
      node = find_node(td->index);
      if(!node)
	fprintf(stderr, "Glass Error: Unable to find node %d referred to in structure tree\n", td->index);
      else
      {
	ntd = (NodeTempData *)node->ad;

	/* Copy the children over */
	ntd->nchildren = td->nchildren;
	for(i = 0; i < td->nchildren; i++)
	  ntd->children[i] = td->children[i];
      }
    }

    t = td->next;
    g_free(td);
    td = t;
  }

  structure_data = NULL;
}

/* Find the surfaces, active points, and variables required by a node */
static void link_nodes(GlassModel *model, GlassNode *node)
{
  NodeTempData *ad;
  GlassTransform *transform;
  GlassSurface *surface;
  GlassActivePoint *apoint;
  GlassVariable *variable;
  int i, nparams, param;

  ad = (NodeTempData *)node->ad;

  /* Link up the surface (if has one) */
  if(strcmp(ad->surface_name->str, ""))
  {
    surface = _glassModelFindSurface(model, ad->surface_name->str);
    if(surface)
      node->surface = surface;
    else
      fprintf(stderr, "Glass Error: Unable to find surface \"%s\"\n", ad->surface_name->str);
  }

  /* Link up the active point (if has one) */
  if(strcmp(ad->apoint_name->str, ""))
  {
    apoint = _glassModelFindActivePoint(model, ad->apoint_name->str);
    if(apoint)
      node->apoint = apoint;
    else
      fprintf(stderr, "Glass Error: Unable to find active point \"%s\"\n", ad->apoint_name->str);
  }

  /* Link up the variables */
  transform = _glassNodeGetTransform(node);
  nparams = _glassTransformGetnParams(_glassTransformGetType(transform));
  for(param = 0; param < nparams; param++)
  {
    /* If has a variable */
    if(strcmp(ad->variable_names[param]->str, ""))
    {
      variable = _glassModelFindVariable(model, ad->variable_names[param]->str);
      if(!variable)
      {
	fprintf(stderr, "Glass Error: Unable to find variable \"%s\"\n", ad->variable_names[param]->str);
	continue;
      }
      _glassTransformSetVariable(transform, param, variable);
    }
  }

  /* Link up the children */
  for(i = 0; i < ad->nchildren; i++)
    _glassNodeAddChild(node, find_node(ad->children[i]));

  /* If a root node put there */
  /* FIXME: This should be in libglassedit too */
  if(ad->is_root)
    _glassModelAddChild(model, node);

  /* Remove the temporary data */
  free_node_temp_data((NodeTempData *)node->ad);
  node->ad = NULL;
}

/* Test if a model has loops in its structure */
/* TODO: Apply this in libglassedit too */
static void test_node_loops(GlassModel *model)
{
  if(test_loop_r(model->children))
  {
    fprintf(stderr, "Loop detected in node structure! All nodes removed.\n");
    _glassFreeNoder(model->children);
    model->children = NULL;
  }
}

/* Recursively test for loops in the node structure */
static boolean test_loop_r(GlassNode *node)
{
  if(!node)
    return FALSE;

  if(node->ad)
    return TRUE;

  node->ad = (void *)TRUE;
  if(test_loop_r(node->next))
  {
    node->ad = NULL;
    return TRUE;
  }
  if(test_loop_r(node->children))
  {
    node->ad = NULL;
    return TRUE;
  }

  return FALSE;
}
