/* 
 * glass - the openGL Articulated Skeleton System
 *
 * Copyright (C) 2001-2003  Robert Ancell <bob27@users.sourceforge.net>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include <stdio.h>
#include <malloc.h>
#include <string.h>

#include "global.h"
#include "glass_material.h"

static void _glassMaterialTestAlpha(GlassMaterial *material);

/* Make a material */
GlassMaterial *_glassMaterialNew(void)
{
  GlassMaterial *material;

  material = (GlassMaterial *)malloc(sizeof(GlassMaterial));
  if(material == NULL)
  {
    fprintf(stderr, "Glass Error: Unable to allocate space for material\n");
    return NULL;
  }
  memset(material, 0, sizeof(GlassMaterial));

  _glassMaterialSetName(material, "");

  /* Default matt grey colour */
  material->ambient[0] = 0.2;
  material->ambient[1] = 0.2;
  material->ambient[2] = 0.2;
  material->ambient[3] = 1.0;
  material->diffuse[0] = 0.8;
  material->diffuse[1] = 0.8;
  material->diffuse[2] = 0.8;
  material->diffuse[3] = 1.0;
  material->specular[3] = 1.0;
  material->emission[3] = 1.0;

  return material;
}

/* Set the material's name */
void _glassMaterialSetName(GlassMaterial *material, const char *name)
{
  if(material->name)
    free(material->name);
  if(name == NULL)
    material->name = strdup("");
  else
    material->name = strdup(name);
}

/* Get the material's name */
const char *_glassMaterialGetName(GlassMaterial *material)
{
  return material->name;
}

/* Set the ambient colour of the material */
void _glassMaterialSetAmbient(GlassMaterial *material,
			      GLfloat r, GLfloat g, GLfloat b, GLfloat a)
{
  material->ambient[0] = r;
  material->ambient[1] = g;
  material->ambient[2] = b;
  material->ambient[3] = a;

  /* Update material->is_alpha */
  _glassMaterialTestAlpha(material);
}

/* Set the ambient colour of the material with an array (vector) */
void _glassMaterialSetAmbientv(GlassMaterial *material,
			       GLfloat *ambient)
{
  material->ambient[0] = ambient[0];
  material->ambient[1] = ambient[1];
  material->ambient[2] = ambient[2];
  material->ambient[3] = ambient[3];

  /* Update material->is_alpha */
  _glassMaterialTestAlpha(material);
}

/* Set the diffuse colour of the material */
void _glassMaterialSetDiffuse(GlassMaterial *material,
			      GLfloat r, GLfloat g, GLfloat b, GLfloat a)
{
  material->diffuse[0] = r;
  material->diffuse[1] = g;
  material->diffuse[2] = b;
  material->diffuse[3] = a;

  /* Update material->is_alpha */
  _glassMaterialTestAlpha(material);
}

/* Set the diffuse colour of the material with an array (vector) */
void _glassMaterialSetDiffusev(GlassMaterial *material,
			       GLfloat *diffuse)
{
  material->diffuse[0] = diffuse[0];
  material->diffuse[1] = diffuse[1];
  material->diffuse[2] = diffuse[2];
  material->diffuse[3] = diffuse[3];

  /* Update material->is_alpha */
  _glassMaterialTestAlpha(material);
}

/* Set the specular reflectance of the material */
void _glassMaterialSetSpecular(GlassMaterial *material,
			       GLfloat r,  GLfloat g,  GLfloat b, GLfloat a)
{
  material->specular[0] = r;
  material->specular[1] = g;
  material->specular[2] = b;
  material->specular[3] = a;

  /* Update material->is_alpha */
  _glassMaterialTestAlpha(material);
}

/* Set the specular reflectance of the material with an array (vector) */
void _glassMaterialSetSpecularv(GlassMaterial *material,
				GLfloat *specular)
{
  material->specular[0] = specular[0];
  material->specular[1] = specular[1];
  material->specular[2] = specular[2];
  material->specular[3] = specular[3];

  /* Update material->is_alpha */
  _glassMaterialTestAlpha(material);
}

/* Set the emmitted colour of the material */
void _glassMaterialSetEmission(GlassMaterial *material,
			       GLfloat r, GLfloat g, GLfloat b, GLfloat a)
{
  material->emission[0] = r;
  material->emission[1] = g;
  material->emission[2] = b;
  material->emission[3] = a;

  /* Update material->is_alpha */
  _glassMaterialTestAlpha(material);
}

/* Set the emmitted colour of the material with an array (vector) */
void _glassMaterialSetEmissionv(GlassMaterial *material,
				GLfloat *emission)
{
  material->emission[0] = emission[0];
  material->emission[1] = emission[1];
  material->emission[2] = emission[2];
  material->emission[3] = emission[3];

  /* Update material->is_alpha */
  _glassMaterialTestAlpha(material);
}

/* Set the shininess of the material */
void _glassMaterialSetShininess(GlassMaterial *material, GLfloat shininess)
{
  material->shininess = shininess;
}

/* Set the texture of the material */
void _glassMaterialSetTexture(GlassMaterial *material, GlassTexture *texture)
{
  material->texture = texture;

  /* Update material->is_alpha */
  _glassMaterialTestAlpha(material);
}

/* Get the texture of the material */
GlassTexture *_glassMaterialGetTexture(GlassMaterial *material)
{
  return material->texture;
}

/* See if the material uses alpha, and set is_alpha if so */
static void _glassMaterialTestAlpha(GlassMaterial *material)
{
  if((material->ambient[3] != 1.0) || (material->diffuse[3] != 1.0))
    material->is_alpha = TRUE;
  else
  {
    if(material->texture != NULL)
      material->is_alpha = material->texture->is_alpha;
    else
      material->is_alpha = FALSE;
  }
}

/* Free the material */
void _glassFreeMaterial(GlassMaterial *material)
{
  if(material->name)
    free(material->name);
  free(material);
}
