/* 
 * glass - the openGL Articulated Skeleton System
 *
 * Copyright (C) 2001-2003  Robert Ancell <bob27@users.sourceforge.net>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include <stdio.h>
#include <malloc.h>

#include "global.h"
#include "glass_node.h"
#include "glass_draw.h"

/* Returns an empty node */
GlassNode *_glassNodeNew(void)
{
  GlassNode *node;

  node = malloc(sizeof(GlassNode));
  if(node == NULL)
  {
    fprintf(stderr, "Glass Error: Unable to allocate space for node\n");
    return NULL;
  }
  memset(node, 0, sizeof(GlassNode));

  _glassNodeSetTransform(node, _glassTransformNew(GLASS_NONE));

  return node;
}

/* Set a nodes transform */
void _glassNodeSetTransform(GlassNode *node, GlassTransform *transform)
{
  node->transform = transform;
}

/* Get a nodes transform */
GlassTransform *_glassNodeGetTransform(GlassNode *node)
{
  return node->transform;
}

/* Set a nodes surface */
void _glassNodeSetSurface(GlassNode *node, GlassSurface *surface)
{
  node->surface = surface;
}

/* Get a nodes surface */
GlassSurface *_glassNodeGetSurface(GlassNode *node)
{
  return node->surface;
}

/* Set a nodes active point */
void _glassNodeSetActivePoint(GlassNode *node, GlassActivePoint *apoint)
{
  node->apoint = apoint;
}

/* Get a nodes active point */
GlassActivePoint *_glassNodeGetActivePoint(GlassNode *node)
{
  return node->apoint;
}

/* Add a child to a node */
void _glassNodeAddChild(GlassNode *node, GlassNode *child)
{
  /* Append */
  if(node->last_child)
  {
    node->last_child->next = child;
    node->last_child = child;
  }
  else
  {
    node->children = child;
    node->last_child = child;
  }
}

/* Recursively work out which nodes are needed for any active points */
int _glassNodeTestActivePointr(GlassNode *node)
{
  int apoint_transform = FALSE;

  if(node == NULL)
    return FALSE;

  /* If this node has an active point, it's important */
  if(_glassNodeGetActivePoint(node) != NULL)
    apoint_transform = TRUE;

  /* If any of this nodes children have active points,
     then this node is important */
  if(_glassNodeTestActivePointr(node->children))
    apoint_transform = TRUE;

  node->apoint_transform = apoint_transform;

  /* Try the next node, irrelevant to this node if it has any
     acitve points, but relevant to the parent of this node */
  if(_glassNodeTestActivePointr(node->next))
    apoint_transform = TRUE;

  return apoint_transform;
}

/* Recursively update the active points */
void _glassNodeUpdateActivePointsr(GlassNode *node)
{
  GlassActivePoint *apoint;

  /* Return if hit the end of a list */
  if(node == NULL)
    return;

  /* Move on to the next node */
  _glassNodeUpdateActivePointsr(node->next);
  
  /* Stop if there are no children of this node with active points */
  if(!node->apoint_transform)
    return;

  glPushMatrix();

  /* Do transform (if it has one) */
  _glassDoTransform(_glassNodeGetTransform(node));

  /* If has an active point, update that */
  if((apoint = _glassNodeGetActivePoint(node)) != NULL)
    _glassActivePointUpdate(apoint);

  /* Move onto next node (that inherits this transform) */
  _glassNodeUpdateActivePointsr(node->children);

  glPopMatrix();
}

/* Free a node */
void _glassFreeNode(GlassNode *node)
{
  _glassFreeTransform(node->transform);
  free(node);
}

/* Free a node, and (recursively) its siblings/children */
void _glassFreeNoder(GlassNode *node)
{
  if(node == NULL)
    return;

  if(node->ad)
    return;

  node->ad = (void *)TRUE;

  _glassFreeNoder(node->next);
  _glassFreeNoder(node->children);

  _glassFreeNode(node);
}
