/* 
 * glass - the openGL Articulated Skeleton System
 *
 * Copyright (C) 2001-2003  Robert Ancell <bob27@users.sourceforge.net>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include <stdio.h>
#include <malloc.h>
#include <string.h>

#include "global.h"
#include "glass_surface.h"
#include "glass_draw.h"

static void _glassFreeSurface(GlassSurface *surface);

/* Returns an empty surface */
GlassSurface *_glassSurfaceNew(const char *name)
{
  GlassSurface *surface;

  surface = (GlassSurface *)malloc(sizeof(GlassSurface));
  if(surface == NULL)
  {
    fprintf(stderr, "Glass Error: Unable to allocate space for surface\n");
    return NULL;
  }
  memset(surface, 0, sizeof(GlassSurface));

  _glassSurfaceSetName(surface, name);

  surface->visible = TRUE;

  return surface;
}

/* Set the name of the surface */
void _glassSurfaceSetName(GlassSurface *surface, const char *name)
{
  if(surface->name)
    free(surface->name);
  if(name == NULL)
    surface->name = strdup("");
  else
    surface->name = strdup(name);
}

/* Get the name of the surface */
const char *_glassSurfaceGetName(GlassSurface *surface)
{
  return surface->name;
}

/* Try to make an openGL display list for the triangles, and one for the alpha triangles */
void _glassSurfaceMakeDisplayLists(GlassSurface *surface)
{
  GLuint index = 0, alpha_index = 0;

  /* If has non-alpha triangles and doesn't yet have a display list */
  if(!surface->index && surface->triangles)
  {
    index = glGenLists(1);
    if(!index)
      return;

    glNewList(index, GL_COMPILE);
      _glassDrawSurface(surface);
    glEndList();

    surface->index = index;
  }

  /* If has alpha triangles and doesn't yet have a display list */
  if(!surface->alpha_index && surface->alpha_triangles)
  {
    alpha_index = glGenLists(1);
    if(!alpha_index)
      return;

    glNewList(alpha_index, GL_COMPILE);
      _glassDrawAlphaSurface(surface);
    glEndList();

    surface->alpha_index = alpha_index;
  }
}

/* Set the visiblility of the surface */
void _glassSurfaceSetVisibility(GlassSurface *surface, int visible)
{
  surface->visible = visible;
}

/* Free a surface */
static void _glassFreeSurface(GlassSurface *surface)
{
  if(surface->name)
    free(surface->name);
  free(surface);
}

/* Free a surface, and the triangles it uses */
void _glassFreeSurfaceChild(GlassSurface *surface)
{
  GlassTriangle *tri_ptr = surface->triangles, *tmp_ptr;

  /* Free the non-alpha triangles */
  while(tri_ptr != NULL)
  {
    tmp_ptr = tri_ptr->next;
    _glassFreeTriangle(tri_ptr);
    tri_ptr = tmp_ptr;
  }

  /* Free the alpha triangles */
  tri_ptr = surface->alpha_triangles;
  while(tri_ptr != NULL)
  {
    tmp_ptr = tri_ptr->next;
    _glassFreeTriangle(tri_ptr);
    tri_ptr = tmp_ptr;
  }

  _glassFreeSurface(surface);
}
